import { AudioPlayerStatus, createAudioPlayer, createAudioResource, entersState, getVoiceConnection, joinVoiceChannel, VoiceConnectionStatus, } from "@discordjs/voice";
import chalk from "chalk";
import { StageChannel } from "discord.js";
import { getAllAudioUrls } from "google-tts-api";
import https from "https";
import ytdl from "ytdl-core";
import { sLogger } from "../logger";
import { shorten } from "../utils/string";
export var VoiceValidateResult;
(function (VoiceValidateResult) {
    VoiceValidateResult[VoiceValidateResult["OK"] = 0] = "OK";
    VoiceValidateResult[VoiceValidateResult["NO_CHANNEL"] = 1] = "NO_CHANNEL";
    VoiceValidateResult[VoiceValidateResult["STAGE_CHANNEL"] = 2] = "STAGE_CHANNEL";
    VoiceValidateResult[VoiceValidateResult["NOT_JOINABLE"] = 3] = "NOT_JOINABLE";
})(VoiceValidateResult || (VoiceValidateResult = {}));
export class VoiceControl {
    handleDisconnect;
    player;
    guildID;
    channelName;
    voiceChannelID;
    isSameChannel(channel) {
        return channel?.id == this.voiceChannelID;
    }
    static validateUser(msg) {
        const channel = msg.member?.voice.channel;
        if (!channel)
            return VoiceValidateResult.NO_CHANNEL;
        if (channel instanceof StageChannel)
            return VoiceValidateResult.STAGE_CHANNEL;
        if (!channel.joinable)
            return VoiceValidateResult.NOT_JOINABLE;
        return VoiceValidateResult.OK;
    }
    constructor(msg, ignorePrivacy, handleDisconnect) {
        this.handleDisconnect = handleDisconnect;
        const validateResult = VoiceControl.validateUser(msg);
        if (validateResult)
            throw validateResult;
        this.player = createAudioPlayer();
        const voiceChannel = msg.member.voice.channel;
        this.guildID = voiceChannel.guild.id;
        this.channelName = voiceChannel.name;
        this.voiceChannelID = voiceChannel.id;
        const connection = joinVoiceChannel({
            channelId: voiceChannel.id,
            guildId: voiceChannel.guild.id,
            adapterCreator: voiceChannel.guild
                .voiceAdapterCreator,
            selfDeaf: !ignorePrivacy,
        });
        connection.subscribe(this.player);
        connection.on(VoiceConnectionStatus.Disconnected, async (_, __) => {
            try {
                await Promise.race([
                    entersState(connection, VoiceConnectionStatus.Signalling, 5000),
                    entersState(connection, VoiceConnectionStatus.Connecting, 5000),
                ]);
            }
            catch (error) {
                sLogger.log(`Disconnected from ${this.channelName}`, "WARNING");
                this.destruct();
            }
        });
    }
    destruct() {
        const connection = getVoiceConnection(this.guildID);
        connection?.destroy();
        this.handleDisconnect();
    }
    async waitTillReady(mode) {
        const connection = getVoiceConnection(this.guildID);
        if (!connection)
            return false;
        try {
            await entersState(connection, VoiceConnectionStatus.Ready, 5000);
            sLogger.log(`Successfully joined ${this.channelName} as ${chalk.magenta(mode)} 🤩🤩`, "SUCCESS");
            return true;
        }
        catch (err) {
            sLogger.log(`Error joining ${this.channelName} : ${err}`, "ERROR");
            return false;
        }
    }
    speakQueue = [];
    async speak(content) {
        sLogger.log(`[TTS] Started Speaking ${shorten(content, 30)} to ${this.channelName}`);
        const results = getAllAudioUrls(content, {
            lang: "th",
            slow: false,
        });
        const initiated = !!this.speakQueue.length;
        this.speakQueue = this.speakQueue.concat(await Promise.all(results.map(async (url) => createAudioResource(await new Promise((res, _) => {
            https.get(url.url, (stream) => {
                res(stream);
            });
        })))));
        if (!initiated)
            this.player.play(this.speakQueue.shift());
        return new Promise((resolve, _) => {
            this.player.on(AudioPlayerStatus.Idle, () => {
                if (this.speakQueue.length) {
                    this.player.play(this.speakQueue.shift());
                }
                else {
                    resolve(true);
                }
            });
            this.player.on("error", (err) => {
                sLogger.log(`[TTS] Error Speaking ${shorten(content, 30)} to ${this.channelName} : ${err}`, "ERROR");
                resolve(false);
            });
        });
    }
    async playSong(url, title, category) {
        const musicStream = ytdl(url, {
            filter: "audioonly",
            quality: "highestaudio",
            highWaterMark: 1 << 25,
            liveBuffer: 4000,
        });
        const musicRc = createAudioResource(musicStream);
        sLogger.log(`[DJCorgi] Playing ${title} in category of ${category}`);
        this.player.play(musicRc);
        return new Promise((resolve, _) => {
            this.player.on(AudioPlayerStatus.Idle, () => {
                resolve(true);
            });
            this.player.on("error", (err) => {
                sLogger.log(`[DJCorgi] Error while playing ${title}: ${err.message}`, "ERROR");
                resolve(false);
            });
        });
    }
    forceSkip() {
        this.player.stop(true);
    }
}
